using System;
using System.Collections.Generic;
using System.Text;
using agsXMPP.protocol.client;
using agsXMPP;
using Fluent.Toc;
using System.Collections;
using System.IO;
using System.Reflection;
using System.Threading;
using System.Xml;
using System.Net;
using XihSolutions.DotMSN;
using XihSolutions.DotMSN.Core;
using XihSolutions.DotMSN.DataTransfer;

namespace ChatBot {
    class ChatBot {
        // These settings are all only for command line stuff!
        public static bool UseCommandLineArgs = false;
        public static Protocol protocol = Protocol.NULL;
        public static String server = null;  // only if you're using Jabber
        public static String user = null;
        public static String pass = null;
        public static String botname = null;

        /// <summary>
        /// Add a Jabber bot to the system
        /// </summary>
        /// <param name="bot">Bot whose message handler you want to use</param>
        /// <param name="server">Server to connect this Jabber bot to.</param>
        /// <param name="user">Username of account</param>
        /// <param name="pass">Password of account</param>
        static void addJabberBot(UserBot bot, String server, String user, String pass) {
            if(server == null)
                throw new ArgumentNullException("Server argument null.  Jabber bot doesn't know what server to connect to.  (Recommendation: jabber.org)");
            if(user == null)
                throw new ArgumentNullException("User argument null.");
            if(pass == null)
                throw new ArgumentNullException("Password argument null.");
            XmppClientConnection xmpp = new XmppClientConnection(server);

            // Open a connection            
            xmpp.Open(user, pass);

            // When logged in, you MUST send your presence!
            xmpp.OnLogin += new ObjectHandler(delegate(object conn) {
                XmppClientConnection x = (XmppClientConnection)conn;
                x.SendMyPresence();
            });
            // Mainly because it's there...might as well use it
            xmpp.OnError += new ErrorHandler(delegate(object sender, Exception ex) {
                Console.WriteLine("Jabber error (" + bot.GetType().Name + "): " + ex.Message);
            });

            // This is what redirects all messages to the bot's message handler
            xmpp.OnMessage += new XmppClientConnection.MessageHandler(bot.ReceivedMessage);

            // Nap time
            while(true) {
                System.Threading.Thread.Sleep(1000);
            }
        }

        static void addJabberBot(UserBot bot, String user, String pass) {
            addJabberBot(bot, "jabber.org", user, pass);
        }

        static void addMSNBot(UserBot bot, String user, String pass) {
            if(user == null)
                throw new ArgumentNullException("User argument null.");
            if(pass == null)
                throw new ArgumentNullException("Password argument null.");

            // Create a Messenger object to use DotMSN.
            XihSolutions.DotMSN.Messenger msn_messenger = new Messenger();

            msn_messenger.NameserverProcessor.ConnectionEstablished += new EventHandler(delegate(object sender, EventArgs e) {
                Console.WriteLine("Connected to server");
            });
            msn_messenger.Nameserver.SignedIn += new EventHandler(delegate(object sender, EventArgs e) {
                NSMessageHandler nsmh = (NSMessageHandler)sender;

                Console.WriteLine("Signed into the messenger network as " + nsmh.Owner.Name);

                AddAllInList(nsmh);
                // set our presence status
                nsmh.Owner.Status = PresenceStatus.Online;
            });
            msn_messenger.Nameserver.SignedOff += new SignedOffEventHandler(delegate(object sender, SignedOffEventArgs e) {
                Console.WriteLine("Signed off from the messenger network");
            });
            msn_messenger.NameserverProcessor.ConnectingException += new ProcessorExceptionEventHandler(delegate(object sender, ExceptionEventArgs e) {
                Console.WriteLine("Connecting failed");
            });
            msn_messenger.Nameserver.ExceptionOccurred += new HandlerExceptionEventHandler(delegate(object sender, ExceptionEventArgs e) {
                Console.WriteLine(e.Exception.ToString() + " Nameserver exception");
            });
            msn_messenger.Nameserver.AuthenticationError += new HandlerExceptionEventHandler(delegate(object sender, ExceptionEventArgs e) {
                Console.WriteLine("Authentication failed");
            });
            msn_messenger.Nameserver.ServerErrorReceived += new ErrorReceivedEventHandler(delegate(object sender, MSNErrorEventArgs e) {
                Console.WriteLine("Server error received");
            });
            msn_messenger.Nameserver.ContactAdded += new ListMutatedAddedEventHandler(delegate(object sender, ListMutateEventArgs e) {
                NSMessageHandler nsmh = (NSMessageHandler)sender;
                AddAllInList(nsmh);
            });

            msn_messenger.ConversationCreated += new ConversationCreatedEventHandler(bot.messenger_ConversationCreated);

            // Emulate an old official microsoft windows messenger client.
            msn_messenger.Credentials.ClientID = "msmsgs@msnmsgr.com";
            msn_messenger.Credentials.ClientCode = "Q1P7W2E4J9R8U3S5";

            if(msn_messenger.Connected) {
                Console.WriteLine("Already connected, so disconnecting from server");
                msn_messenger.Disconnect();
            }

            // set the credentials, this is ofcourse something every DotMSN program will need to
            // implement.
            msn_messenger.Credentials.Account = user;
            msn_messenger.Credentials.Password = pass;

            // Inform the user what is happening and try to connect to the messenger network.			
            Console.WriteLine("Connecting to server");
            msn_messenger.Connect();
        }

        private static void AddAllInList(NSMessageHandler nsmh) {
            foreach(Contact contact in nsmh.ContactList.All) {
                if(!contact.OnAllowedList) {
                    contact.OnAllowedList = true;
                }
            }
        }

        /// <summary>
        /// Add an AIM bot to the system
        /// </summary>
        /// <param name="bot">Bot whose message handler you want to use</param>
        /// <param name="user">Username of account</param>
        /// <param name="pass">Password of account</param>
        static void addAIMBot(UserBot bot, String user, String pass) {
            if(user == null)
                throw new ArgumentNullException("User argument null.");
            if(pass == null)
                throw new ArgumentNullException("Password argument null.");

            TocClient tc = new TocClient();
            // Add that message handler again
            tc.Message += new MessageEventHandler(bot.ReceivedMessage);
            // Toc has a lot less console feedback, so this is nice.
            Console.WriteLine("Signing on");
            try {
                tc.SignOn(user, pass);
                tc.Error += new Fluent.Toc.ErrorEventHandler(delegate(object sender, Fluent.Toc.ErrorEventArgs e) {
                    Console.WriteLine("AIM error (" + bot.GetType().Name + "): " + e.Message);
                });
                tc.StartListening();
            } catch(TocException e) {
                Console.WriteLine("AIM error (" + bot.GetType().Name + "): " + e.Message);
            }

            // More nap time
            while(true) {
                System.Threading.Thread.Sleep(1000);
            }
        }

        static void addBots() {
            Assembly assembly = Assembly.GetCallingAssembly();
            // This is where we iterate through every class in my assembly and check for things that inherit UserBot
            Type UserBotT = Type.GetType("ChatBot.UserBot");

            // This is if you pass in a bot name via the command line only
            // Pretty spiffy huh?  Reflection is cool.
            if(botname != null) {
                foreach(Type type in assembly.GetTypes()) {
                    if(type.IsClass) {
                        if(type.IsSubclassOf(UserBotT) && type.Name == botname) {
                            if(protocol == Protocol.AIM) {
                                addAIMBot((UserBot)Activator.CreateInstance(type), user, pass);
                                return;
                            } else if(protocol == Protocol.Jabber) {
                                if(server == null) {
                                    throw new ArgumentException("No server specified for Jabber bot.");
                                }
                                addJabberBot((UserBot)Activator.CreateInstance(type), server, user, pass);
                                return;
                            } else if(protocol == Protocol.MSN) {
                                Console.WriteLine("Adding MSN Bot");
                                addMSNBot((UserBot)Activator.CreateInstance(type), user, pass);
                                return;
                            } else {
                                throw new ArgumentNullException("Protocol argument null.");
                            }
                        }
                    }
                }
            } else {
                Hashtable usernames = new Hashtable(32);

                foreach(Type type in assembly.GetTypes()) {
                    if(type.IsClass && type.IsSubclassOf(UserBotT) &&
                        type.Name != "UserBot" && type.Name != "DialogBot" &&
                        type.Name != "BasicBot") {
                        UserBot tempBot = (UserBot)Activator.CreateInstance(type);
                        Console.WriteLine(type.Name + ": Use value is " + tempBot.Use);

                        // Only use the bot if it's specified to do so
                        if(!tempBot.Use)
                            continue;

                        String[] loginInfo = tempBot.GetLoginInfo();

                        // Only use the bot if we haven't already logged into the
                        // account that it uses.
                        if(usernames.ContainsKey(loginInfo[0]))
                            continue;
                        usernames.Add(loginInfo[0], true);

                        // Make sure the loginInfo string is valid                        
                        Thread t = new Thread(delegate() {
                            if(tempBot.BotProtocol == Protocol.Jabber) {
                                if(loginInfo.Length == 3) {
                                    addJabberBot(tempBot, loginInfo[2], loginInfo[0], loginInfo[1]);
                                } else if(loginInfo.Length == 2) {
                                    addJabberBot(tempBot, loginInfo[0], loginInfo[1]);
                                } else {
                                    throw new InvalidDataException("Error: Trying to initialize " + type.Name + " as a Jabber bot, but required login information not given!  Expecting 2 arguments (user, password) or 3 arguments (server, user, password).");
                                }
                            } else if(tempBot.BotProtocol == Protocol.AIM) {
                                if(loginInfo.Length != 2) {
                                    throw new InvalidDataException("Error: Trying to initialize " + type.Name + " as a AIM bot, but required login information not given!  Expecting 2 arguments (user, password).");
                                }
                                addAIMBot(tempBot, loginInfo[0], loginInfo[1]);
                            } else if(tempBot.BotProtocol == Protocol.MSN) {
                                Console.WriteLine("Connecting to MSN...");
                                addMSNBot(tempBot, loginInfo[0], loginInfo[1]);
                            }


                        });
                        t.Start();
                    }
                }
            }

        }

        /// <summary>
        /// Handles command line arguments.  Should be pretty self-explanatory
        /// </summary>
        /// <param name="args">Arguments array from Main</param>
        static void ParseCommandLineArguments(String[] args) {
            int lastPos = args.Length - 1;
            for(int i = 0; i < args.Length; i++) {
                switch(args[i]) {
                    case "-j":
                        if(protocol != Protocol.NULL) {
                            throw new ArgumentException("Already defined protocol.");
                        }
                        protocol = Protocol.Jabber;
                        break;
                    case "-m":
                        if (protocol != Protocol.NULL)
                        {
                            throw new ArgumentException("Already defined protocol.");
                        }
                        protocol = Protocol.MSN;
                        break;
                    case "-a":
                        if(protocol != Protocol.NULL) {
                            throw new ArgumentException("Already defined protocol.");
                        }
                        protocol = Protocol.AIM;
                        break;
                    case "-u":
                        if(user != null) {
                            throw new ArgumentException("Already defined username.");
                        }
                        if(i == lastPos) {
                            throw new ArgumentException("No username specified.");
                        }
                        user = args[++i];
                        break;
                    case "-p":
                        if(pass != null) {
                            throw new ArgumentException("Already defined password.");
                        }
                        if(i == lastPos) {
                            throw new ArgumentException("No password specified.");
                        }
                        pass = args[++i];
                        break;
                    case "-s":
                        if(server != null) {
                            throw new ArgumentException("Already defined server.");
                        }
                        if(i == lastPos) {
                            throw new ArgumentException("No server specified.");
                        }
                        server = args[++i];
                        break;
                    case "-b":
                        if(botname != null) {
                            throw new ArgumentException("Can only launch one bot from the command line, sorry.");
                        }
                        if(i == lastPos) {
                            throw new ArgumentException("No bot specified.");
                        }
                        botname = args[++i];
                        break;
                }
            }
        }

        static void Main(string[] args) {
            ParseCommandLineArguments(args);
            ChatBot.addBots();

            // Keep this process alive so the program doesn't terminate
            while(true) {
                System.Threading.Thread.Sleep(1000);
            }
        }

        public enum Protocol {
            NULL,
            Jabber,
            AIM,
            MSN
        }
    }
}
