using System;

using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using agsXMPP.protocol.client;

namespace ChatBot
{
    class ConversationTree
    {
        public object sender;
        public ConversationNode rootNode, currentNode = null;
        private LinkedList<ConversationNode> anywhereNodes;
        private RegexOptions options;

        public ConversationTree(ConversationNode rootNode) : this(rootNode, RegexOptions.IgnoreCase & RegexOptions.Multiline) {}

        public ConversationTree(ConversationNode rootNode, RegexOptions yourOptions)
        {
            this.rootNode = rootNode;            
            anywhereNodes = new LinkedList<ConversationNode>();
            options = yourOptions;
        }

        public void addUniversalNode(ConversationNode node){
            anywhereNodes.AddLast(node);
        }

        public void addUniversalNode(ConversationNode node, Regex rex)
        {
            node.regularExpression = rex;
            anywhereNodes.AddLast(node);
        }

        /**
         * Finds the next node that This should go to and goes to it.  Returns true only
         * if a successful match is made.
         */
        public bool findNext(UserMessage message)
        {
	    // Starting out: assign root to current
            if(currentNode == null) {
                currentNode = rootNode;
                return true;
            }

            Match m;
                        
            // first, check universal nodes
            foreach(ConversationNode uni in anywhereNodes) {
                if(uni.regularExpression == null) {
                    throw new Exception("Error: universal nodes must have a regular expression!");
                }
                m = uni.regularExpression.Match(message.Message);
                if(m.Success) {
                    currentNode = uni;
                    return true;
                }
            }

            // second, check children nodes
            ConversationNode cn;
            Regex rex;
            ConversationNode.ShouldIDisplayHandler sh;
            foreach(Object[] child in currentNode.UserResponses) {
                //m = Regex.Match(message, child.regexExpression, options);
                cn = (ConversationNode)child[0];
                
                if(child[1] == null) {
                    currentNode = cn;
                    return true;
                }

                if(child[1] is Regex) {
                    rex = (Regex)child[1];
                    Console.WriteLine("Regex is " + rex.ToString());
                    Console.WriteLine();
                    m = rex.Match(message.Message);
                    if(m.Success) {
                        currentNode = cn;
                        return true;
                    }
                } else if(child[1] is ConversationNode.ShouldIDisplayHandler){
                    sh = (ConversationNode.ShouldIDisplayHandler)child[1];
                    Console.WriteLine("Processing ShouldIDisplayHandler");
                    if(sh(sender, message)) {
                        currentNode = cn;
                        return true;
                    }
                }
            }

            // third...doesn't match anything.  Say as much.
            return false;
        }

        public void resetToRoot()
        {
            currentNode = rootNode;
        }

        internal class ConversationNode
        {
            public LinkedList<Object[]> UserResponses;
            //public String regexExpression;
            public Regex regularExpression;
            public String botResponse;
            public delegate void InvocationHandler(object sender, UserMessage message);
            private InvocationHandler myFunc;
            public delegate bool ShouldIDisplayHandler(object sender, UserMessage message);

            // For univeral nodes only
            /*public ConversationNode(String expression, String botResponse)
            {
                universal = true;
                UserResponses = new LinkedList<Object[]>();
                this.botResponse = botResponse;
            }*/

            public ConversationNode(String botResponse)
            {
                //universal = false;
                UserResponses = new LinkedList<Object[]>();
                this.botResponse = botResponse;
            }
            public ConversationNode()
            {
                //universal = false;
                UserResponses = new LinkedList<Object[]>();
            }
            /*
            public ConversationNode(InvocationHandler ih)
            {
                universal = false;
                UserResponses = new LinkedList<ConversationNode>();
                setFunction(ih);
            }
            */

            

            public void addChild(ConversationNode cn, Regex rex){
                /*if(cn.universal) {
                    throw new ArgumentException("Conversation node is universal!  Can't be childed.");
                }*/
                Object[] ConversationPairing = { cn, rex };
                UserResponses.AddLast(ConversationPairing);
            }

            public void addChild(ConversationNode cn, ShouldIDisplayHandler e)
            {
                /*if(cn.universal) {
                    throw new ArgumentException("Conversation node is universal!  Can't be childed.");
                }*/
                Object[] ConversationPairing = { cn, e };
                UserResponses.AddLast(ConversationPairing);
            }

            public void addChild(ConversationNode cn)
            {
                /*if(cn.universal) {
                    throw new ArgumentException("Conversation node is universal!  Can't be childed.");
                }*/
                Object[] ConversationPairing = { cn, null };
                UserResponses.AddLast(ConversationPairing);
            }

            public String getResponse(){
                if(botResponse == "" || botResponse == null) {
                    throw new Exception("Bot Response string is empty!  Make sure you're running setFunction if you don't give the bot response in the constructor, and that it sets a bot response.");
                }
                return botResponse;
            }
            public void setResponse(String botResponse)
            {
                this.botResponse = botResponse;
            }

            public void setFunction(InvocationHandler ih)
            {
                myFunc = ih;
            }

            public bool hasFunction()
            {
                return myFunc != null;
            }

            public void invoke(object sender, UserMessage m)
            {
                if(hasFunction()) myFunc(sender, m);
            }
        }
    }   
}
