using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using Message = agsXMPP.protocol.client.Message;
using MessageEventArgs = Fluent.Toc.MessageEventArgs;
using System.Text.RegularExpressions;
using Node = ChatBot.ConversationTree.ConversationNode;
using ChatBot;
using Fluent.Toc;
using agsXMPP;
using agsXMPP.protocol.client;
using XihSolutions.DotMSN;
using XihSolutions.DotMSN.Core;


namespace ChatBot
{
    // Parent class of all bots -- not for modification
    abstract class UserBot
    {
	
        // Default login info for Jabber.
        protected string jabber_im_username = "nanodeathbot";
        protected string jabber_im_password = "nuked";
        protected string jabber_im_server = "jabber.org";

        // Default login info for AIM.
        protected string aim_im_username = "relkin0bot";
        protected string aim_im_password = "nuked";

        //protected virtual string aim_im_username = "relkin0bot2";
        //protected virtual string aim_im_password = "nuked";

        // Default login for MSN.
        protected string msn_username = "youthslam@hotmail.com";
        protected string msn_password = "nfbslam";

        // Keep users' conversations separate
        private Hashtable conversations = new Hashtable();

	  // This is the state variable.
	  public static string state = "0";

        // Should this bot be used.  Default set to true.
        public virtual Boolean Use
        {
            get
            {
                return true;
            }
        }

        public virtual ChatBot.Protocol BotProtocol
        {
            get
            {
                // Bot Protocol defaults to MSN.
                return ChatBot.Protocol.MSN;
            }
        }

        public void ReceivedMessage(object sender, object message)
        {
            // Message is a Jabber message;
            // MessageEventArgs is an AIM message.
            // TextMessageEventArgs is a MSN message.
            // Both get handled here!
            if(message is Message) {
                Message m = (Message)message;
                GotMessage(sender, new UserMessage(m.From.Bare, m.Body, false, message));
            } else if(message is MessageEventArgs) {
                MessageEventArgs m = (MessageEventArgs)message;
                GotMessage(sender, new UserMessage(m.From, m.Message, m.AutoResponse, message));
            } else if (message is TextMessageEventArgs) {
                TextMessageEventArgs m = (TextMessageEventArgs)message;
                GotMessage(sender, new UserMessage(m.Sender.Name, m.Message.Text, false, message));
            }
        }

        //  Message handling helper function particular to MSN.
        public void messenger_ConversationCreated(object sender, ConversationCreatedEventArgs e)
        {
            // Check if the request is initiated remote or by this object
            // if it is initiated remote then we have to create a conversation form. Otherwise the 
            // form is already created and we don't need to create another one.
            //NSMessageHandler nsmh = (NSMessageHandler)sender;
            //Console.WriteLine("Sender name: " + nsmh.Owner.Name);
            if (e.Initiator == null)
            {
                Console.WriteLine("Message received: " + e.Conversation.Messenger.Owner.Name);

                e.Conversation.Switchboard.TextMessageReceived +=
                    new TextMessageReceivedEventHandler(ReceivedMessage);
            }
        }

        // Reponsible for generating the dialog tree.
        public abstract ConversationTree GenerateTree();

        // Simple interface that takes in a message and a username and returns a response.
        public virtual string HandleMessage(string message, string user)
        {
            return "Boring default message.  You need to override HandleMessage!";
        }

        // Return the server, username, and password of this chat bot.
        public virtual String[] GetLoginInfo()
        {
            if (BotProtocol == ChatBot.Protocol.Jabber) {
                return new String[] { jabber_im_username, jabber_im_password, jabber_im_server };
            } else if (BotProtocol == ChatBot.Protocol.MSN) {
                return new String[] { msn_username, msn_password };
            } else {
                return new String[] { aim_im_username, aim_im_password };
            }
        }


        /// <summary>
        /// When a message is received, it gets converted to a UserMessage and then sent here.
        /// This generates the conversation tree if necessary and carries out the necessary logic
        /// involving tree traversal and message-sending.
        /// </summary>
        /// <param name="Sender"></param>
        /// <param name="message"></param>
        public void GotMessage(object Sender, UserMessage message)
        {
            // This makes sure there's one and only one conversation tree per user talking to this bot
            ConversationTree ct;
            if(conversations.Contains(message.From)) {
                ct = (ConversationTree)conversations[message.From];
            } else {
                ct = GenerateTree();
                ct.sender = Sender;
                conversations[message.From] = ct;
            }
            // Guaranteed to have a tree by this point

            // Acknowledge in console
            Console.WriteLine(this.GetType().Name + " got message! Message was " + message.Message);

            // Move the user to the next node based on the message
            ct.findNext(message);

            // Poke that node
            ct.currentNode.invoke(Sender, message);

            // Send the message containing the node's response
            SendMessageNow(Sender, ct.currentNode.getResponse(), message);         
        }

        /// <summary>
        /// Send a message now!
        /// </summary>
        /// <param name="sender">Sender object that got passed to ReceivedMessage (one of the protocol clients)</param>
        /// <param name="YourMessage">What the bot should say</param>
        /// <param name="originalMessage">The original user message, so the destination is known</param>
        public void SendMessageNow(object sender, String YourMessage, UserMessage originalMessage)
        {
            switch(BotProtocol) {
                case ChatBot.Protocol.AIM:
                    TocClient tc = (TocClient)sender;
                    tc.Send(originalMessage.From, YourMessage);
                    break;
                case ChatBot.Protocol.Jabber:
                    XmppClientConnection xmpp = (XmppClientConnection)sender;
                    xmpp.Send(new Message(((Message)originalMessage.Original).From, MessageType.chat, YourMessage));
                    break;
                case ChatBot.Protocol.MSN:
                    SBMessageHandler sbmh = (SBMessageHandler)sender;
                    TextMessage tmessage = new TextMessage(YourMessage);
                    sbmh.SendTextMessage(tmessage);
                    break;
            }
        }

    }
    /// <summary>
    /// Uniform handling of messages on both protocols
    /// </summary>
    class UserMessage
    {
        private String from;
        public String From {
            get
            {
                return from;
            }
        }

        private String message;
        public String Message
        {
            get
            {
                return message;
            }
        }

        private Boolean autoresponse;
        public Boolean AutoResponse
        {
            get
            {
                // I don't think this is actually used, but AIM requires it or something
                // Just kept around in case
                return autoresponse;
            }
        }

        private Object original;
        public Object Original
        {
            get
            {
                // Original message is stored, just in case
                return original;
            }
        }

        public UserMessage(String From, String Message, Boolean AutoResponse, Object Original)
        {
            try {
                from = From;
                message = Regex.Replace(Message, @"<(.|\n)*?>", string.Empty); // AIM sends HTML!
                autoresponse = AutoResponse;
                original = Original;
            } catch {

            }
        }
    }
}
